#include "Tag.h"
#include "Exception.h"
#include "Log.h"
#include "__mysql.h"
#include "__for.h"
#include "__if.h"
#include "__read.h"
#include "__set.h"
#include "__jet.h"
#include "__while.h"
#include <iostream>

namespace jet {
   
   Tag::Tag(coreutils::ZString &in, coreutils::MString &parent)
     : ZString(in), parent(parent) {
	if(in.ifNext("<")) {
	   name = in.getTokenInclude("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_!");	   
	   if(in.startsWith(" ") || in.startsWith("/") || in.startsWith(">")) {
	      coreutils::Log(coreutils::LOG_DEBUG_1) << "Processing tag: " << name;
	      bool finished = false;
	      while(!finished) {
		 in.skipWhitespace();
		 if(in.ifNext(">")) {
		    hasContainer = true;
		    finished = true;
		    break;
		 } else if(in.ifNext("/>")) {
		    hasContainer = false;
		    finished = true;
		    break;
		 }
		 if(!finished) {
		    coreutils::ZString keywordName = in.getTokenInclude("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_");
		    if(in.ifNext("=\"")) {
		       keywords[keywordName] = in.getTokenExclude("\"");
		    }
		    if(!in.ifNext("\"")) {}
		 }
	      }
	      if(hasContainer) {
		 coreutils::Log(coreutils::LOG_DEBUG_2) << "has Container: " << hasContainer;
		 char *start = in.getCursor();
		 while(!in.eod()) {
		    char *end = in.getCursor();
		    if(ifEndTagName(in)) {
		       container = coreutils::ZString(start, end - start);
		       coreutils::Log(coreutils::LOG_DEBUG_2) << "Container: [" << container << "]";
		       break;		       
		    } else if(ifNested(in)) {}
		    else
		      in.nextChar();
		 }		 
	      }
	      setZString(in.parsed());
	      if(hasContainer && evaluate)
		parseContainer(container);
	   }
	} else
	  throw coreutils::Exception("First character of (in) must be a <.");
     }
   
   Tag::~Tag() {
      if(evaluate) {	   
	 		copyContainer(out, parent);
      } else {
	 		copyContainer(container, parent);	 
      }      
   }
   
   void Tag::parseContainer(coreutils::ZString &in) {
   	char *start = in.getCursor();
      while(!in.eod()) {
			if(in.startsWith("<")) {
		   	if(ifTagName(in, "mysql")) {
	   	   	__mysql _mysql(in, out);
	  	   	 	continue;
	   		} else if(ifTagName(in, "for")) {
	      		__for _for(in, out);
	      		continue;
	   		} else if(ifTagName(in, "if")) {
	      		__if _if(in, out);
	      		continue;
	   		} else if(ifTagName(in, "jet")) {
	      		__jet _jet(in, out);
	      		continue;
	   		} else if(ifTagName(in, "read")) {
	      		__read _read(in, out);
	      		continue;
	   		} else if(ifTagName(in, "set")) {
	      		__set _set(in, out);
	      		continue;
	   		} else if(ifTagName(in, "while")) {
	      		__while _while(in, out);
	      		continue;
	   		}
	 		}
	 		out.write(in.charAt(0));
	 		in.nextChar();
      }
   }
   
   void Tag::scanContainer(coreutils::ZString &in) {
      while(!in.eod()) {
	 if(ifEndTagName(in)) 
	   return;
	 else if(ifNested(in)) {}	 
	 else in.nextChar();
      }
   }
   
   bool Tag::ifTagName(coreutils::ZString &in, const char *tag) {
      in.push();
      if(in.ifNext("<"))
        if(in.getTokenInclude("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_").equals(tag)) {       
	   if(in.ifNext(" ")) {
	      in.pop();
	      return true;
	   } else if(in.ifNext("/>")) {
	      in.pop();
	      return true;
	   } else if(in.ifNext(">")) {
	      in.pop();
	      return true;
	   }	  
	}
      in.pop();
      return false;
   }
   
   bool Tag::ifTagName(coreutils::ZString &in) {
      in.push();
      if(in.ifNext("<"))
        if(in.getTokenInclude("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_").equals(name)) {
	   in.push();
	   if(in.ifNext(" ")) {
	      in.pop();
	      return true;
	   } else if(in.ifNext("/>")) {
	      in.pop();
	      return true;
	   } else if(in.ifNext(">")) {
	      in.pop();
	      return true;
	   }	   
	   in.pop();
	}
      in.pop();
      return false;
   }
   
   bool Tag::ifEndTagName(coreutils::ZString &in) {
      in.push();
      if(in.ifNext("</"))
        if(in.getTokenInclude("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_").equals(name)) {
	   if(in.ifNext(">")) 
	     return true;
	}      
      in.pop();
      return false;
   }
   
   int Tag::skipBlankLine(coreutils::ZString in) {
      ZString temp = in.getTokenInclude(" \t");
      if(ifNext("\n"))
        return temp.getLength() + 1;
      return 0;
   }
   
   void Tag::copyContainer(coreutils::ZString &in, coreutils::MString &out) {
      while(!in.eod()) {	 
         if(filterBlankLines) {
				in.push();
				if(!in.eod()) {
					in.getTokenInclude(" \t");
					if(!in.ifNext("\n")) {
						in.pop();
						// TODO: shouod be safe to output line here.
					}
				}
	 		}
         else {     
	    		out.write(in.charAt(0));
	    		in.nextChar();
	 		}	 
      }
   }
   
   bool Tag::keywordExists(coreutils::ZString keyword) {
      return keywords.find(keyword) != keywords.end();
   }
   
   bool Tag::ifNested(coreutils::ZString &in) {
      bool hasContainer = false;
      if(ifTagName(in)) {
	 while(!in.eod()) {
	    in.skipWhitespace();
	    if(in.ifNext(">")) {
	       hasContainer = true;
	       break;
	    } else if(in.ifNext("/>")) {
	       hasContainer = false;
	       return true;
	    } else if(in.getTokenExclude("=").getLength() != 0) {
	       if(in.ifNext("=\"")) {
		  while(1) {
		     if(in.ifNext("\"")) {
			break;
		     }
		     in.nextChar();
		  }
	       }
	    }
	 }
	 if(hasContainer)
	   scanContainer(in);
      }
      return false;
   }
}
